unit wdxNTFSFileStreams;

interface

uses
  JclWin32, JclSysInfo, JclFileUtils, JclSecurity, JclNTFS,
  SysUtils,
  Classes,
  Windows,
  FSPlugin,
  StrUtils;

function MakePath(S: string): string;
function Occurrences(const Substring, Text: string): Integer; inline;

function FsContentGetSupportedField(FieldIndex: Integer; FieldName: PAnsiChar;
  Units: PAnsiChar; MaxLen: Integer): Integer; stdcall;
function FsContentGetValueW(FileName: PWideChar; FieldIndex, UnitIndex: Integer;
  FieldValue: PByte; MaxLen, Flags: Integer): Integer; stdcall;
function FsContentGetValue(FileName: PAnsiChar; FieldIndex, UnitIndex: Integer;
  FieldValue: PByte; MaxLen, Flags: Integer): Integer; stdcall;
function FsContentGetDefaultSortOrder(FieldIndex: Integer): Integer; stdcall;
function FsContentGetDefaultView(ViewContents, ViewHeaders, ViewWidths,
  ViewOptions: PAnsiChar; MaxLen: Integer): BOOL; stdcall;
function FsContentGetDefaultViewW(ViewContents, ViewHeaders, ViewWidths,
  ViewOptions: PWideChar; MaxLen: Integer): BOOL; stdcall;

var
  UsedwdxStreamIds: TStreamIds;
  msg_Dir, msg_Stream: string;

function LoadIconWithScaleDown(hinst: HINST; pszName: LPCWSTR; cx: Integer;
  cy: Integer; var phico: HICON): HResult; stdcall;

implementation

const
  FieldsNames: array [0..4] of PAnsiChar = ('Has Stream', 'Streams Count',
    'Stream', 'All Streams One Line', 'Type');
  FieldsTypes: array[0..High(FieldsNames)] of Integer = (6, 1, 11, 11, 11);

const
  DetectStr = 'MULTIMEDIA';

function LoadIconWithScaleDown; external 'comctl32.dll' name 'LoadIconWithScaleDown' delayed;

function Occurrences(const Substring, Text: string): Integer; inline;
var
  offset: Integer;
begin
  Result := 0;
  offset := PosEx(Substring, Text, 1);
  while offset <> 0 do
  begin
    inc(result);
    offset := PosEx(Substring, Text, offset + length(Substring));
  end;
end;

function GetVeryLongFileNameForce(const FullFileName: string): string; inline;
begin
  Result := FullFileName;
  //if Length(FullFileName) < 256 then Exit; Force
  if Pos('\\?\', FullFileName) > 0 then Exit;

  if Pos('\\', FullFileName) > 0 then
    Result := '\\?\UNC\' + Copy(FullFileName, 3, Length(FullFileName) - 2)
  else
    Result := '\\?\' + FullFileName;
end;

function DropStreamNameAttr(const StreamStr: string): string;
var
  Str: string;
  APos: Integer;
begin
  Result := StreamStr;

  Str := UpperCase(Result);
  APos := Pos(':$DATA', Str);

  if (APos > 0) and (APos  = Length(Result) - Length(':$DATA') + 1) then
    Result := Copy(Result, 1, APos - 1);

  if Result[1] = ':' then
    Result := Copy(Result, 2, Length(Result));
end;

procedure FillStringsFromData(const Data: TFindStreamData;
  const StreamNumber: Integer; var Counter: Integer;
  var LastStreamStr, StreamNumberedStr, StreamsAllStr: string);
begin
  LastStreamStr := '';

  if Data.Name <> '' then
  begin
    LastStreamStr := DropStreamNameAttr(Data.Name);
  end;
  Inc(Counter);

  if (Counter > 0) and (StreamNumber = Counter) then
    StreamNumberedStr := LastStreamStr;

  if LastStreamStr <> '' then
    if StreamsAllStr <> '' then
      StreamsAllStr := StreamsAllStr + '; ' + LastStreamStr
    else
      StreamsAllStr := LastStreamStr;
end;

function GetStreamsForFile(const FileName: string; const StreamNumber: Integer;
   var StreamsCount: Integer; var StreamNumberedStr, StreamsAllStr: string): Boolean;
var
  Data: TFindStreamData;
  Counter: Integer;
  LastStreamStr: string;
  StreamIds: TStreamIds;
begin
  Result := False;
  Counter := 0;
  StreamsCount := 0;
  StreamNumberedStr := '';
  StreamsAllStr := '';
  StreamIds := UsedwdxStreamIds;

  ZeroMemory(@Data, SizeOf(Data));
  Data.Internal.FileHandle := INVALID_HANDLE_VALUE;

  if NtfsFindFirstStream(GetVeryLongFileNameForce(FileName), StreamIds, Data) then
  try
    FillStringsFromData(Data, StreamNumber, Counter,
      LastStreamStr, StreamNumberedStr, StreamsAllStr);

    Result := True;

    while NtfsFindNextStream(Data) do
    begin
      FillStringsFromData(Data, StreamNumber, Counter,
        LastStreamStr, StreamNumberedStr, StreamsAllStr);
    end;

  finally
    NtfsFindStreamClose(Data);
  end;

  StreamsCount := Counter;
end;

procedure ContentGetDetectString(DetectString: PAnsiChar; MaxLen: integer); stdcall;
begin
  StrLCopy(DetectString, PAnsiChar(DetectStr), MaxLen - 1);
end;

function FsContentGetSupportedField(FieldIndex: Integer; FieldName: PAnsiChar;
  Units: PAnsiChar; MaxLen: Integer): Integer; stdcall;
begin
  if FieldIndex in [0 .. High(FieldsNames)] then
  begin
    if FieldIndex = 2 then
      StrLCopy(Units, '1|2|3|4|5|6|7|8|9|10', MaxLen - 1)
    else
      StrCopy(Units, '');
    StrLCopy(FieldName, FieldsNames[FieldIndex], MaxLen - 1);
    Result := FieldsTypes[FieldIndex];
  end
  else
    Result := ft_NoMoreFields;
end;

function MakePath(S: string): string;
begin
  Result := S;

  if S[1] <> '\' then
    Exit;

  if Length(S) < 2 then
    Exit
  else
  begin
    //Result := S;
    //Result[1] := Result[2];
    //Result[2] := ':';
    Result := Copy(S, 2, Length(S));
  end;

  //MessageBox(GetActiveWindow, PChar(S + #13 + Result), nil, MB_ICONINFORMATION);
end;

function GetItemType(FileName: string): string;
  function CountSubStr(const C: Char; const S: string): Integer;
  var
    X: Integer;
  begin
    Result := 0;
    for X := 1 to Length(S) do
      Inc(Result, Ord(S[X] = C));
  end;

  function CountSubStrEx(const C: Char; const MoreThan: Integer;
    const S: string): Boolean;
  var
    X, Counter: Integer;
  begin
    Result := False;
    Counter := 0;
    for X := 1 to Length(S) do
    begin
      Inc(Counter, Ord(S[X] = C));
      if Counter > MoreThan then
      begin
        Result := True;
        Break;
      end;
    end;
  end;

  function StringByType(Directory: Boolean; FN: string): string;
  begin
    if Directory then
      Result := msg_Dir
    else
    begin
      Result := Copy(ExtractFileExt(FN), 2, Length(FN));
      //Result := ExtractFileExt(FN);
      // if Result = '' then
      // Result := '.';
    end;
  end;

var
  AFileName: string;
begin
  Result := '';

  AFileName := ExcludeTrailingPathDelimiter(FileName);

  if CountSubStrEx(':', 1, AFileName) then
    Result := msg_Stream
  else if SysUtils.FileExists(AFileName) then
    Result := StringByType(False, AFileName)
  else if SysUtils.DirectoryExists(AFileName) then
    Result := StringByType(True, AFileName)
  else if JclFileUtils.DirectoryExists(AFileName) then
    Result := StringByType(True, AFileName)
  else if JclFileUtils.FileExists(AFileName) then
    Result := StringByType(False, AFileName)
  else
    Result := '?';
end;

function FsContentGetValueW(FileName: PWideChar; FieldIndex, UnitIndex: integer; FieldValue: PByte;
  MaxLen, Flags: integer): integer; stdcall;
var
  AFileName: string;
  StreamNumberedStr, StreamsAllStr: string;
  StreamsCount, NumberStream: Integer;
begin
  //AFileName := GetVeryLongFileName(FileName);
  AFileName := MakePath(FileName);
  Result := ft_fieldempty;

  if AFileName[1] = ':' then
    Exit;

  NumberStream := UnitIndex;
  Inc(NumberStream);

  GetStreamsForFile(AFileName, NumberStream, StreamsCount, StreamNumberedStr, StreamsAllStr);

  case FieldIndex of
    0: PInteger(FieldValue)^ := Integer(StreamsCount > 0);
    1: PInteger(FieldValue)^ := StreamsCount;
    2: StrLCopy(PChar(FieldValue), PChar(StreamNumberedStr), MaxLen div 2 - 2);
    3: StrLCopy(PChar(FieldValue), PChar(StreamsAllStr), MaxLen div 2 - 2);
    4: StrLCopy(PChar(FieldValue), PChar(GetItemType(AFileName)), MaxLen div 2 - 2);
  end; //case
  if FieldIndex in [0..High(FieldsNames)] then
    Result := FieldsTypes[FieldIndex]
    else Result := ft_fieldempty;
end;

function FsContentGetValue(FileName:pansichar;FieldIndex,UnitIndex:integer;FieldValue:pbyte;
  maxlen,flags:integer):integer; stdcall;
begin
  Result := FsContentGetValueW(PChar(WideString(AnsiString(FileName))),
    FieldIndex, UnitIndex, FieldValue, maxlen, flags);
end;

function FsContentGetDefaultSortOrder(FieldIndex: integer): integer; stdcall;
begin
  case FieldIndex of
    4:
      Result := 1;
  else
    Result := -1;
  end;
end;

function FsContentGetDefaultView(ViewContents, ViewHeaders, ViewWidths,
  ViewOptions: PAnsiChar; MaxLen: Integer): BOOL; stdcall;
begin
  Result := False;
end;

function FsContentGetDefaultViewW(ViewContents, ViewHeaders, ViewWidths,
  ViewOptions: PWideChar; MaxLen: Integer): BOOL; stdcall;
begin
  Result := True;
  StrLCopy(ViewContents,
    PChar('[=<fs>.Type]\n[=<fs>.Has Stream]\n[=<fs>.Streams Count]\n[=<fs>.All Streams One Line]'),
    //PChar('[=tc.size.bkMGT2]\n[=<fs>.Type]\n[=<fs>.Has Stream]\n[=<fs>.Streams Count]\n[=<fs>.All Streams in One Line]'),
    MaxLen div 2 - 2);
  StrLCopy(ViewHeaders,
    PChar('Type\nHas Stream\nStreams Count\nAll Streams One Line'),
    //PChar('Size\nType\nHas Stream\nStreams Count\nAll Streams One Line'),
    MaxLen div 2 - 2);
  StrLCopy(ViewWidths,
    PChar('80,-30,-40,-37,50,510'),
    //PChar('80,-30,-40,-40,-37,50,510'),
    MaxLen div 2 - 2);
  StrLCopy(ViewOptions,
    PChar('-1|0'),
    MaxLen div 2 - 2);
end;

end.
