library wfx_FileStreams;

{

  2005. Alexeev Alexander. All Rights Reserved.
 WFX plugin for Total Commander (5.51 and later).

 Required: fsplugin.pas and JEDI Code Library (JCL)
 
}

uses
  SysConst, Windows, SysUtils, Classes, Registry, IniFiles, ShellAPI, JclWin32,
  JclSysInfo, JclFileUtils, JclDateTime, JclSecurity, JclNTFS, FSPlugin,
  UnitProperties;

{$E wfx}

{$R *.res}

{__ Types _____________________________________________________________________}

type
  PSearchResult = ^TSearchResult;
  TSearchResult = record
    ItemType: Integer;
    Path: String;
    CallNo: Integer;
  end;

  PSearchResult_Root = ^TSearchResult_Root;
  TSearchResult_Root = record
    GSR: TSearchResult;
    Drives: String;
  end;

  PSearchResult_Dir = ^TSearchResult_Dir;
  TSearchResult_Dir = record
    GSR: TSearchResult;
    Handle: THandle;
  end;

  PSearchResult_File = ^TSearchResult_File;
  TSearchResult_File = record
    GSR: TSearchResult;
    Data: TFindStreamData;
    ftCreationTime: TFileTime;
    ftLastAccessTime: TFileTime;
    ftLastWriteTime: TFileTime;
  end;

{__ Constants _________________________________________________________________}

const
  CopyBufferSize    = 655360;
  itRoot            = 1;
  itDir             = 2;
  itFile            = 3;
  AllowedAttributes = FILE_ATTRIBUTE_READONLY or FILE_ATTRIBUTE_HIDDEN or
                      FILE_ATTRIBUTE_SYSTEM or FILE_ATTRIBUTE_DIRECTORY or
                      FILE_ATTRIBUTE_ARCHIVE or FILE_ATTRIBUTE_DEVICE or
                      FILE_ATTRIBUTE_NORMAL or FILE_ATTRIBUTE_TEMPORARY or
                      FILE_ATTRIBUTE_COMPRESSED or FILE_ATTRIBUTE_OFFLINE or
                      FILE_ATTRIBUTE_NOT_CONTENT_INDEXED or FILE_ATTRIBUTE_ENCRYPTED;

{__ Global variables __________________________________________________________}

var
  Plug_ID: Integer;
  ProgressProc: TProgressProc;
  LogProc: TLogProc;
  RequestProc: TRequestProc;
  InitFailed: Boolean;
  Reg: TRegIniFile;
  StreamTypes: TStreamIds;

{__ Language support __________________________________________________________}

  msg_MkDirUnsupported,
  msg_ExecuteUnsupported,
  msg_NotImplemented,
  msg_RootCopyUnsupported,
  msg_ExecuteFailed,
  msg_ReadError,
  msg_DeleteError,
  msg_WriteError: String;

{__ Support routines __________________________________________________________}

procedure SetDefaultLang;
begin
msg_MkDirUnsupported:='You cann''t create directories inside files.';
msg_ExecuteUnsupported:='You cann''t execute file streams.';
msg_NotImplemented:='Not implemented yet.';
msg_RootCopyUnsupported:='Copying to root directory isn''t supported.';
msg_ExecuteFailed:='Cann''t run program.';
msg_ReadError:='Error reading file or stream.';
msg_DeleteError:='Error deleting file or stream.';
msg_WriteError:='Error writing file or stream.';

msg_PropCaption:='';
msg_PropOk:='';
msg_PropCancel:='';
msg_PropBrowse:='';
msg_PropStreams:='';
msg_PropOpenTitle:='';
msg_PropOpenFilter:='';
msg_PropLangNotFound:='Specified language file was not found.';
msg_PropLangInvalidFile:='Specified language file for plugin is invalid.';
end;

procedure FillFindData_Root (var FindData: TWIN32FINDDATA; Drive: Char);
begin
FindData.dwFileAttributes:=FILE_ATTRIBUTE_DIRECTORY or FILE_ATTRIBUTE_NORMAL;
FindData.ftCreationTime:=DateTimeToFileTime (Now);
FindData.ftLastAccessTime:=FindData.ftCreationTime;
FindData.ftLastWriteTime:=FindData.ftCreationTime;
FindData.nFileSizeHigh:=0;
FindData.nFileSizeLow:=0;
FindData.cFileName [0]:=Drive;
FindData.cAlternateFileName [0]:=Drive;
end;

procedure FillFindData_Stream (var FindData: TWIN32FINDDATA; const StreamData: TFindStreamData);
var
  S: String;

begin
FindData.dwFileAttributes:=FILE_ATTRIBUTE_NORMAL;
if (StreamData.Attributes and STREAM_MODIFIED_WHEN_READ) <> 0 then
 FindData.dwFileAttributes:=FindData.dwFileAttributes or FILE_ATTRIBUTE_READONLY;
if (StreamData.Attributes and STREAM_CONTAINS_SECURITY) <> 0 then
 FindData.dwFileAttributes:=FindData.dwFileAttributes or FILE_ATTRIBUTE_SYSTEM;
FindData.nFileSizeHigh:=StreamData.Size shr 32;
FindData.nFileSizeLow:=StreamData.Size;
S:=StreamData.Name;
if S = '' then
 begin
 if StreamData.StreamID = siStandard then
  S:='::$DATA'
 else
  begin
  FindData.dwFileAttributes:=FindData.dwFileAttributes or FILE_ATTRIBUTE_HIDDEN;
  case StreamData.StreamID of
   siExtendedAttribute:
     S:='::$EA_DATA';
   siSecurity:
     S:='::$SECURITY_DESCRIPTOR';
   siAlternate:
     S:='::$ALTERNATE_DATA';
   siHardLink:
     S:='::$LINK';
   siProperty:
     S:='::$PROPERTY_DATA';
   siObjectIdentifier:
     S:='::$OBJECT_ID';
   siReparsePoints:
     S:='::$REPARSE_DATA';
   siSparseFile:
     S:='::$SPARSE_BLOCK';
  else // siInvalid:
   S:='::';
  end;
  end;
 end;
Move (S [1], FindData.cFileName [0], Length (S));
end;

function  MakePath (S: String): String;
begin
if Length (S) < 2 then
 Result:=S
else
 begin
 Result:=S;
 Result [1]:=Result [2];
 Result [2]:=':';
 end;
end;

function  Min (X, Y: Integer): Integer;
begin
if X > Y then
 Result:=Y
else
 Result:=X;
end;

const
  UNC = '\\?\';

function FileExists (const FileName: string): Boolean;
var
  hFile: THandle;
  WS: WideString;

begin
if (Pos ('..', FileName) > 0) or ((Length (FileName) <= 3){ and (FileName [2] = ':')}) then
 begin
 Result:=False;
 Exit;
 end;
WS:=UNC+FileName;
hFile:=CreateFileW (PWideChar (WS), 0, FILE_SHARE_DELETE or FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);
Result:=hFile <> INVALID_HANDLE_VALUE;
if Result then
 CloseHandle (hFile);
end;

function  DirectoryExists (const Directory: string): Boolean;
var
  Code: Integer;
  WS: WideString;

begin
WS:=UNC+IncludeTrailingPathDelimiter (Directory);
Code:=GetFileAttributesW (PWideChar (WS));
Result:=(Code <> -1) and (FILE_ATTRIBUTE_DIRECTORY and Code <> 0);
end;

function ForceDirectories (Dir: String): Boolean;
var
  E: EInOutError;
  WS: WideString;

begin
Result:=True;
if Dir = '' then
 begin
 E:=EInOutError.CreateRes (@SCannotCreateDir);
 E.ErrorCode:=3;
 raise E;
 end;
Dir:=ExcludeTrailingPathDelimiter (Dir);
if (Length (Dir) < 3) or DirectoryExists (Dir) or (ExtractFilePath (Dir) = Dir) then
 Exit;
WS:=UNC+IncludeTrailingPathDelimiter (Dir);
Result:=ForceDirectories (ExtractFilePath (Dir)) and CreateDirectoryW (PWideChar (WS), nil);
end;

function DeleteDirectory (const DirectoryName: String): Boolean;
var
  WS: WideString;

begin
WS:=UNC+IncludeTrailingPathDelimiter (DirectoryName);
Result:=RemoveDirectoryW (PWideChar (WS));
end;

function  DeleteFile (const FileName: String): Boolean;
var
  WS: WideString;

begin
if DirectoryExists (FileName) then
 begin
 Result:=DeleteDirectory (FileName);
 Exit;
 end;
WS:=UNC+FileName;
Result:=Windows.DeleteFileW (PWideChar (WS));
end;

function  DelTreeEx (const Path: string; AbortOnFailure: Boolean; Progress: TDelTreeProgress): Boolean;
var
  Files: TStringList;
  LPath: String;
  FileName: String;
  I: Integer;
  PartialResult: Boolean;
  Attr: DWORD;

begin
if Path = '' then
 begin
 Result:=False;
 Exit;
 end;
Result:=True;
Files:=TStringList.Create;
try
 LPath:=PathRemoveSeparator (Path);
 BuildFileList (LPath+'\*.*', faAnyFile, Files);
 for I:=0 to Files.Count-1 do
  begin
  FileName:=LPath+'\'+Files [I];
  PartialResult:=True;
  // If the current file is itself a directory then recursively delete it
  Attr:=GetFileAttributes (PChar (FileName));
  if (Attr <> DWORD (-1)) and ((Attr and FILE_ATTRIBUTE_DIRECTORY) <> 0) then
   PartialResult:=DelTreeEx (FileName, AbortOnFailure, Progress)
  else
   begin
   if Assigned (Progress) then
    PartialResult:=Progress (FileName, Attr);
   if PartialResult then
    begin
    // Set attributes to normal in case it's a readonly file
    PartialResult:=SetFileAttributes (PChar (FileName), FILE_ATTRIBUTE_NORMAL);
    if PartialResult then
     PartialResult:=DeleteFile (FileName);
    end;
   end;
  if not PartialResult then
   begin
   Result := False;
   if AbortOnFailure then
    Break;
   end;
  end;
finally
 FreeAndNil (Files);
end;
if Result then
 begin
 // Finally remove the directory itself
 Result:=SetFileAttributes (PChar (LPath), FILE_ATTRIBUTE_NORMAL);
 if Result then
  Result:=DeleteDirectory (LPath);
 end;
end;

function DelTree (const Path: string): Boolean;
begin
Result:=DelTreeEx (Path, False, nil);
end;

function  ParseStreamName (RemoteName: PChar; out Name: String): Boolean;
var
  X: Integer;

begin
Name:=RemoteName;
X:=Pos (':', Name);
if X > 0 then
 begin
 if (Length (Name) > 8) and (AnsiUpperCase (Copy (Name, Length (Name)-6, 7)) = '::$DATA') then
  begin
  Name:=ExcludeTrailingPathDelimiter (MakePath (Copy (Name, 1, Length (Name)-7)));
  Result:=False;
  end
 else
  begin
  Delete (Name, X-1, 1);
  Name:=ExcludeTrailingPathDelimiter (MakePath (Name));
  Result:=True;
  end;
 end
else
 begin
 Name:=ExcludeTrailingPathDelimiter (MakePath (Name));
 Result:=False;
 end;
end;

procedure LoadSetup;
var
  Ini: TIniFile;
  LangFile: String;
  Streams: Integer;
  Str: TStringList;

begin
SetDefaultLang;
Streams:=-1;
try
 if TC_Ver.DefaultIniName <> '' then
  Ini:=TIniFile.Create (TC_Ver.DefaultIniName)
 else
  Ini:=TIniFile.Create ('TC_wfx_NTFSFileStreams.ini');
 try
  LangFile:=Ini.ReadString ('wfx_NTFSFileStreams', 'Language', '');
  Streams:=Ini.ReadInteger ('wfx_NTFSFileStreams', 'StreamTypes', Streams);
  if (LangFile <> '') and FileExists (LangFile) then
   begin
   Str:=TStringList.Create;
   try
    Str.LoadFromFile (LangFile);
    if Str [0] = LangFileHeader then
     begin
     msg_MkDirUnsupported:=Str [2];
     msg_ExecuteUnsupported:=Str [3];
     msg_NotImplemented:=Str [4];
     msg_RootCopyUnsupported:=Str [5];
     msg_PropCaption:=Str [6];
     msg_PropOk:=Str [7];
     msg_PropCancel:=Str [8];
     msg_PropBrowse:=Str [9];
     msg_PropStreams:=Str [10];
     msg_PropOpenTitle:=Str [11];
     msg_PropOpenFilter:=Str [12];
     msg_PropLangNotFound:=Str [13];
     msg_PropLangInvalidFile:=Str [14];
     msg_ExecuteFailed:=Str [15];
     msg_ReadError:=Str [16];
     msg_DeleteError:=Str [17];
     msg_WriteError:=Str [18];
     end;
   finally
    FreeAndNil (Str);
   end;
   end;
 finally
  FreeAndNil (Ini);
 end;
except
 on E: Exception do;
end;
StreamTypes:=[];
if Streams and Flag_siInvalid <> 0 then
 Include (StreamTypes, siInvalid);
if Streams and Flag_siStandard <> 0 then
 Include (StreamTypes, siStandard);
if Streams and Flag_siExtendedAttribute <> 0 then
 Include (StreamTypes, siExtendedAttribute);
if Streams and Flag_siSecurity <> 0 then
 Include (StreamTypes, siSecurity);
if Streams and Flag_siAlternate <> 0 then
 Include (StreamTypes, siAlternate);
if Streams and Flag_siHardLink <> 0 then
 Include (StreamTypes, siHardLink);
if Streams and Flag_siProperty <> 0 then
 Include (StreamTypes, siProperty);
if Streams and Flag_siObjectIdentifier <> 0 then
 Include (StreamTypes, siObjectIdentifier);
if Streams and Flag_siReparsePoints <> 0 then
 Include (StreamTypes, siReparsePoints);
if Streams and Flag_siSparseFile <> 0 then
 Include (StreamTypes, siSparseFile);
end;

function FsFindClose (Hdl: THandle): Integer; stdcall; forward;
function FsDeleteFile (RemoteName: PChar): Bool; stdcall; forward;

{__ Plugin routines ___________________________________________________________}

function FsInit (PluginNr: Integer; pProgressProc: TProgressProc; pLogProc:
    TLogProc; pRequestProc: TRequestProc): Integer; stdcall;
begin
try
 Plug_ID:=PluginNr;
 ProgressProc:=pProgressProc;
 LogProc:=pLogProc;
 RequestProc:=pRequestProc;
 InitFailed:=(Win32Platform <> VER_PLATFORM_WIN32_NT) or (not
     EnableProcessPrivilege (True, SE_BACKUP_NAME) or (not
     EnableProcessPrivilege (True, SE_RESTORE_NAME)));
 if InitFailed then
  begin
  Result:=0;
  Exit;
  end;
 try
  Reg:=TRegIniFile.Create ('\', KEY_READ);
  Reg.RootKey:=HKEY_CLASSES_ROOT;
 except
  on E: Exception do
   Reg:=nil;
 end;
 LoadSetup;
 Result:=0;
except
 on E: Exception do
  begin
  InitFailed:=True;
  Result:=1;
  end;
end;
end;

function FsFindFirst (path: PChar; var FindData: TWIN32FINDDATA): THandle; stdcall;
var
  S, G: String;
  I, X: Integer;
  C: Char;
  SR: PSearchResult_Root;
  SD: PSearchResult_Dir;
  SF: PSearchResult_File;
  Drives: packed array [0..25] of Boolean;
  Handle: THandle;
  StreamData: TFindStreamData;
  SRec: TSearchRec;

begin
Result:=0;
try
 ZeroMemory (@FindData, SizeOf (TWIN32FINDDATA));
 if InitFailed then
  begin
  Result:=INVALID_HANDLE_VALUE;
  SetLastError (ERROR_NO_MORE_FILES);
  Exit;
  end;
 if (Path = nil) or (Path = '\') then
  begin
  X:=0;
  SetLength (G, 513);
  ZeroMemory (@(Drives [0]), Length (Drives));
  for C:='A' to 'Z' do
   begin
   S:=C+':';
   ZeroMemory (@(G [1]), Length (G));
   QueryDosDevice (PChar (S), PChar (G), 512);
   if (GetDriveType (PChar (S)) = DRIVE_FIXED) and (LastDelimiter (':?', G) <= 0)
       and (GetVolumeFileSystem (C) = 'NTFS') then
    begin
    Inc (X);
    Drives [Ord (C)-Ord ('A')]:=True;
    end;
   end;
  Finalize (G);
  Finalize (S);
  SetLength (S, X);
  I:=1;
  for X:=0 to Length (Drives) do
   if Drives [X] then
    begin
    S [I]:=Chr (Ord ('A')+X);
    Inc (I);
    end;
  if Length (S) <= 0 then
   begin
   Result:=INVALID_HANDLE_VALUE;
   SetLastError (ERROR_NO_MORE_FILES);
   Exit;
   end;
  FillFindData_Root (FindData, S [1]);
  GetMem (SR, SizeOf (TSearchResult_Root));
  ZeroMemory (SR, SizeOf (TSearchResult_Root));
  Result:=THandle (SR);
  SR^.GSR.ItemType:=itRoot;
  SR^.GSR.Path:=Path;
  SR^.GSR.CallNo:=0;
  SR^.Drives:=S;
  Finalize (S);
  end
 else
  begin
  if Length (Path) < 2 then
   begin
   Result:=INVALID_HANDLE_VALUE;
   SetLastError (ERROR_NO_MORE_FILES);
   Exit;
   end;
  S:=ExcludeTrailingPathDelimiter (MakePath (Path));
  if FileExists (S) then
   begin
   ZeroMemory (@StreamData, SizeOf (TFindStreamData));
   if not NtfsFindFirstStream (S, StreamTypes, StreamData) then
    begin
    Result:=INVALID_HANDLE_VALUE;
    SetLastError (ERROR_NO_MORE_FILES);
    Exit;
    end;
   FillFindData_Stream (FindData, StreamData);
   GetMem (SF, SizeOf (TSearchResult_File));
   ZeroMemory (SF, SizeOf (TSearchResult_File));
   Result:=THandle (SF);
   SF^.GSR.ItemType:=itFile;
   SF^.GSR.Path:=Path;
   SF^.GSR.CallNo:=0;
   SF^.Data:=StreamData;
   if GetFileInformation (S, SRec) then
    begin
    {$WARNINGS OFF}
    SF^.ftCreationTime:=SRec.FindData.ftCreationTime;
    SF^.ftLastAccessTime:=SRec.FindData.ftLastAccessTime;
    SF^.ftLastWriteTime:=SRec.FindData.ftLastWriteTime;
    {$WARNINGS ON}
    end;
   FindData.ftCreationTime:=SF^.ftCreationTime;
   FindData.ftLastAccessTime:=SF^.ftLastAccessTime;
   FindData.ftLastWriteTime:=SF^.ftLastWriteTime;
   end
  else
   begin
   S:=S+'\*.*';
   Handle:=Windows.FindFirstFile (PChar (S), FindData);
   if Handle = INVALID_HANDLE_VALUE then
    Exit;
   FindData.dwFileAttributes:=(FindData.dwFileAttributes and AllowedAttributes) or FILE_ATTRIBUTE_DIRECTORY;
   GetMem (SD, SizeOf (TSearchResult_Dir));
   ZeroMemory (SD, SizeOf (TSearchResult_Dir));
   Result:=THandle (SD);
   SD^.GSR.ItemType:=itDir;
   SD^.GSR.Path:=Path;
   SD^.GSR.CallNo:=0;
   SD^.Handle:=Handle;
   end;
  end;
except
 on E: Exception do
  begin
  if Result <> 0 then
   FsFindClose (Result);
  Result:=INVALID_HANDLE_VALUE;
  SetLastError (ERROR_GEN_FAILURE);
  end;
end;
end;

function FsFindNext (Hdl: THandle; var FindData: TWIN32FINDDATA): Bool; stdcall;
var
  S: PSearchResult;
  SR: PSearchResult_Root;
  SD: PSearchResult_Dir;
  SF: PSearchResult_File;

begin
try
 ZeroMemory (@FindData, SizeOf (TWIN32FINDDATA));
 if InitFailed or (Hdl = 0) then
  begin
  Result:=False;
  Exit;
  end;
 S:=PSearchResult (Hdl);
 Inc (S^.CallNo);
 case S^.ItemType of
  itRoot:
    begin
    SR:=PSearchResult_Root (Hdl);
    if S^.CallNo < Length (SR^.Drives) then
     begin
     FillFindData_Root (FindData, SR^.Drives [S^.CallNo+1]);
     Result:=True;
     end
    else
     Result:=False;
    end;

  itDir:
    begin
    SD:=PSearchResult_Dir (Hdl);
    Result:=Windows.FindNextFile (SD^.Handle, FindData);
    FindData.dwFileAttributes:=(FindData.dwFileAttributes and AllowedAttributes) or FILE_ATTRIBUTE_DIRECTORY;
    end;

  itFile:
    begin
    SF:=PSearchResult_File (Hdl);
    Result:=NtfsFindNextStream (SF^.Data);
    if Result then
     begin
     FillFindData_Stream (FindData, SF^.Data);
     FindData.ftCreationTime:=SF^.ftCreationTime;
     FindData.ftLastAccessTime:=SF^.ftLastAccessTime;
     FindData.ftLastWriteTime:=SF^.ftLastWriteTime;
     end;
    end;
 else
  Result:=False;
 end;
except
 on E: Exception do
  Result:=False;
end;
end;

function FsFindClose (Hdl: THandle): Integer; stdcall;
var
  S: PSearchResult;
  SR: PSearchResult_Root;
  SD: PSearchResult_Dir;
  SF: PSearchResult_File;

begin
try
 if InitFailed or (Hdl = 0) or (Hdl = INVALID_HANDLE_VALUE) then
  begin
  Result:=0;
  Exit;
  end;
 S:=PSearchResult (Hdl);
 case S^.ItemType of
  itRoot:
    begin
    SR:=PSearchResult_Root (Hdl);
    Finalize (SR^.Drives);
    end;

  itDir:
    begin
    SD:=PSearchResult_Dir (Hdl);
    if SD^.Handle <> INVALID_HANDLE_VALUE then
     Windows.FindClose (SD^.Handle);
    end;

  itFile:
    begin
    SF:=PSearchResult_File (Hdl);
    NtfsFindStreamClose (SF^.Data);
    end;
 end;
 Finalize (S^.Path);
 FreeMem (S);
except
 on E: Exception do;
end;
Result:=0;
end;

function FsMkDir (RemoteDir: PChar): Bool; stdcall;
var
  S: String;

begin
if InitFailed then
 begin
 Result:=True;
 Exit;
 end;
try
 S:=ExcludeTrailingPathDelimiter (MakePath (RemoteDir));
 if FileExists (ExcludeTrailingPathDelimiter (ExtractFilePath (S))) then
  begin
  RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_MkDirUnsupported), nil, 0);
  Result:=False;
  Exit;
  end;
 Result:=ForceDirectories (S);
except
 on E: Exception do
  Result:=False;
end;
end;

function FsExecuteFile (MainWin: THandle; RemoteName, Verb: PChar): Integer; stdcall;
var
  S, Cmd, Args: String;
  shex: TShellExecuteInfo;
  X: Integer;

begin
if InitFailed then
 begin
 Result:=FS_EXEC_OK;
 Exit;
 end;
try
 S:=AnsiLowerCase (Verb);
 if RemoteName = '\' then
  begin
  if (S = 'properties') and CallProperties (MainWin) then
   LoadSetup;
  Result:=FS_EXEC_OK;
  Exit;
  end;
 if S = 'open' then
  begin
  if Length (RemoteName) < 2 then
   begin
   Result:=FS_EXEC_ERROR;
   Exit;
   end;
  if ParseStreamName (RemoteName, S) then
   begin
   RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_ExecuteUnsupported), nil, 0);
   Result:=FS_EXEC_ERROR;
   Exit;
   end;
  if ShellExecute (MainWin, 'open', PChar (S), nil, PChar (ExtractFilePath (S)), SW_NORMAL) <= 32 then
   begin
   RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_ExecuteFailed), nil, 0);
   Result:=FS_EXEC_ERROR;
   end
  else
   Result:=FS_EXEC_OK;
  (*
     isn't working???
  Move (S [1], RemoteName [0], Min (Length (S)+1, MAX_PATH));
  Result:=FS_EXEC_SYMLINK;
  *)
  end
 else
 if S = 'properties' then
  begin
  if Length (RemoteName) < 2 then
   begin
   Result:=FS_EXEC_ERROR;
   Exit;
   end;
  if ParseStreamName (RemoteName, S) then
   begin
   RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_NotImplemented), nil, 0);
   Result:=FS_EXEC_OK;
   Exit;
   end;
  ZeroMemory (@shex, SizeOf (shex));
  shex.cbSize:=SizeOf (shex);
  shex.fMask:=SEE_MASK_INVOKEIDLIST;
  shex.nShow:=SW_SHOW;
  shex.Wnd:=MainWin;
  shex.lpVerb:='properties';
  shex.lpFile:=PChar (S);
  if ShellExecuteEx (@shex) then
   Result:=FS_EXEC_OK
  else
   begin
   RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_ExecuteFailed+': '+SysErrorMessage (GetLastError)), nil, 0);
   Result:=FS_EXEC_ERROR;
   end;
  end
 else
  begin
  S:=Copy (S, 1, Pos (' ', S)-1);
  if S = 'chmod' then
   begin
   RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_NotImplemented), nil, 0);
   Result:=FS_EXEC_ERROR;
   end
  else
  if S = 'quote' then
   begin
   S:=Verb;
   S:=Copy (S, Pos (' ', S)+1, MAXINT);
   X:=Pos (' ', S);
   if X > 0 then
    begin
    Cmd:=AnsiLowerCase (Copy (S, 1, X-1));
    Args:=Trim (Copy (S, X+1, MaxInt));
    if (Args <> '') and (Args [1] = '"') then
     Args:=Copy (Args, 2, MaxInt);
    if (Args <> '') and (Args [Length (Args)] = '"') then
     Args:=Copy (Args, 1, Length (Args)-1);
    if (Cmd = 'del') or (Cmd = 'rmdir') then
     if FileExists (Args) then
      begin
      ChDir (MakePath (RemoteName));
      if not DeleteFile (Args) then
       RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_DeleteError+': '+SysErrorMessage (GetLastError)), nil, 0);
      Result:=FS_EXEC_OK;
      Exit;
      end
     else
      begin
      ChDir (MakePath (RemoteName));
      if not DeleteDirectory (Args) then
       RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_DeleteError+': '+SysErrorMessage (GetLastError)), nil, 0);
      Result:=FS_EXEC_OK;
      Exit;
      end
    else
     if Cmd = 'mkdir' then
      begin
      ChDir (MakePath (RemoteName));
      if not ForceDirectories (Args) then
       RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_WriteError+': '+SysErrorMessage (GetLastError)), nil, 0);
      Result:=FS_EXEC_OK;
      Exit;
      end;
    end;
   if ShellExecute (MainWin, 'open', PChar (S), nil, PChar (MakePath (RemoteName)), SW_NORMAL) <= 32 then
    begin
    RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_ExecuteFailed), nil, 0);
    Result:=FS_EXEC_ERROR;
    end
   else
    Result:=FS_EXEC_OK;
   end
  else
   Result:=FS_EXEC_ERROR;
  end;
except
 on E: Exception do
  Result:=FS_EXEC_ERROR;
end;
end;

function FsGetFile (RemoteName, LocalName: PChar; CopyFlags: Integer;
    RemoteInfo: pRemoteInfo): Integer; stdcall;
var
  LF, RF: String;
  Sz, Sz100, Offset: Int64;
  SzLo: DWord;
  hFile, hMapS, hMapD: THandle;
  Source, Destination: Pointer;
  HiSz, HiOffset: PDWord;
  WS: WideString;

 function  CopyDataBlock (BlockSize: Integer): Integer;
 begin
 Result:=0;
 Source:=MapViewOfFile (hMapS, FILE_MAP_READ, HiOffset^, Offset, BlockSize);
 if Source = nil then
  begin
  Result:=FS_FILE_READERROR;
  Exit;
  end;
 try
  Destination:=MapViewOfFile (hMapD, FILE_MAP_WRITE, HiOffset^, Offset, BlockSize);
  if Destination = nil then
   begin
   Result:=FS_FILE_WRITEERROR;
   Exit;
   end;
  try
   Move (Source^, Destination^, BlockSize);
  finally
   UnmapViewOfFile (Destination);
  end;
 finally
  UnmapViewOfFile (Source);
 end;
 Sz:=Sz-BlockSize;
 Offset:=Offset+BlockSize;
 if ProgressProc (Plug_ID, RemoteName, LocalName, Offset div Sz100) = 1 then
  begin
  Result:=FS_FILE_USERABORT;
  Exit;
  end;
 end;

begin
Result:=FS_FILE_OK;
if InitFailed then
 Exit;
try
 // prepare copy
 LF:=LocalName;
 if FileExists (LF) and ((CopyFlags and FS_COPYFLAGS_OVERWRITE) = 0) then
  begin
  Result:=FS_FILE_EXISTS;
  Exit;
  end;
 ParseStreamName (RemoteName, RF);
 if ProgressProc (Plug_ID, RemoteName, LocalName, 0) = 1 then
  begin
  Result:=FS_FILE_USERABORT;
  Exit;
  end;
 // open files
 SetLastError (0);
 WS:=UNC+RF;
 hFile:=CreateFileW (PWideChar (WS), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);
 if hFile = INVALID_HANDLE_VALUE then
  begin
  RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_ReadError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
  Result:=FS_FILE_OK;
  Exit;
  end;
 try
  Sz:=0;
  HiSz:=Pointer (DWord (@Sz)+4);
  SzLo:=Windows.GetFileSize (hFile, HiSz);
  if (SzLo = INVALID_FILE_SIZE) and (GetLastError <> NO_ERROR) then
   begin
   Result:=FS_FILE_READERROR;
   Exit;
   end;
  Sz:=Sz+SzLo;
  if Sz > 0 then
   begin
   SetLastError (0);
   hMapS:=CreateFileMapping (hFile, nil, PAGE_READONLY, HiSz^, SzLo, nil);
   end;
 finally
  CloseHandle (hFile);
 end;
 if (Sz > 0) and (hMapS = 0) then
  begin
  RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_ReadError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
  Result:=FS_FILE_OK;
  Exit;
  end;
 try
  WS:=UNC+LF;
  hFile:=CreateFileW (PWideChar (WS), GENERIC_READ or GENERIC_WRITE, 0, nil, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN or FILE_ATTRIBUTE_NORMAL, 0);
  if hFile = INVALID_HANDLE_VALUE then
   begin
   Result:=FS_FILE_WRITEERROR;
   Exit;
   end;
  try
   if Sz > 0 then
    hMapD:=CreateFileMapping (hFile, nil, PAGE_READWRITE, HiSz^, SzLo, nil);
  finally
   CloseHandle (hFile);
  end;
  if (Sz > 0) and (hMapD = 0) then
   begin
   Result:=FS_FILE_WRITEERROR;
   Exit;
   end;
  try
   if Sz > 0 then
    begin
    // copy data
    Sz100:=Sz div 100;
    if Sz100 <= 0 then
     Sz100:=1;
    HiOffset:=Pointer (DWord (@Offset)+4);
    Offset:=0;
    while Sz >= CopyBufferSize do
     begin
     Result:=CopyDataBlock (CopyBufferSize);
     if Result <> 0 then
      Exit;
     end;
    if Sz <> 0 then
     begin
     Result:=CopyDataBlock (Sz);
     if Result <> 0 then
      Exit;
     end;
    end;
   // close files
  finally
   if hMapD <> 0 then
    CloseHandle (hMapD);
  end;
 finally
  if hMapS <> 0 then
   CloseHandle (hMapS);
 end;
 // finish job
 if ProgressProc (Plug_ID, RemoteName, LocalName, 100) = 1 then
  begin
  Result:=FS_FILE_USERABORT;
  Exit;
  end;
 if (CopyFlags and FS_COPYFLAGS_MOVE) <> 0 then
  if FsDeleteFile (RemoteName) then
   Result:=FS_FILE_OK
  else
   begin
   RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_DeleteError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
   Result:=FS_FILE_OK;
   end
 else
  Result:=FS_FILE_OK
except
 on E: Exception do
  Result:=FS_FILE_READERROR;
end;
end;

function FsPutFile (LocalName, RemoteName: PChar; CopyFlags: Integer): Integer; stdcall;

 function  CountSubStr (C: Char; const S: String): Integer;
 var
   X: Integer;

 begin
 Result:=0;
 for X:=1 to Length (S) do
  Inc (Result, Ord (S [X] = C));
 end;

var
  LF, RF, PossibleLocation, UploadAs, VPath: String;
  Sz, Sz100, Offset: Int64;
  SzLo: DWord;
  hFile, hMapS, hMapD: THandle;
  Source, Destination: Pointer;
  HiSz, HiOffset: PDWord;
  X: Integer;
  WS: WideString;

 function  CopyDataBlock (BlockSize: Integer): Integer;
 begin
 Result:=0;
 Source:=MapViewOfFile (hMapS, FILE_MAP_READ, HiOffset^, Offset, BlockSize);
 if Source = nil then
  begin
  Result:=FS_FILE_READERROR;
  Exit;
  end;
 try
  Destination:=MapViewOfFile (hMapD, FILE_MAP_WRITE, HiOffset^, Offset, BlockSize);
  if Destination = nil then
   begin
   Result:=FS_FILE_WRITEERROR;
   Exit;
   end;
  try
   Move (Source^, Destination^, BlockSize);
  finally
   UnmapViewOfFile (Destination);
  end;
 finally
  UnmapViewOfFile (Source);
 end;
 Sz:=Sz-BlockSize;
 Offset:=Offset+BlockSize;
 if ProgressProc (Plug_ID, LocalName, PChar (RF), Offset div Sz100) = 1 then
  begin
  Result:=FS_FILE_USERABORT;
  Exit;
  end;
 end;

begin
Result:=FS_FILE_OK;
if InitFailed then
 Exit;
try
 // prepare copy
 RF:=RemoteName;
 X:=LastDelimiter ('\', RF);
 if X <= 1 then
  begin
  RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_RootCopyUnsupported), nil, 0);
  Result:=FS_FILE_NOTSUPPORTED;
  Exit;
  end;
 LF:=LocalName;
 if not FileExists (LF) then
  begin
  Result:=FS_FILE_NOTFOUND;
  Exit;
  end;
 VPath:=Copy (RF, 1, X);
 RF:=MakePath (RF);
 UploadAs:=Copy (RF, X+1, MaxInt);
 if (Pos (':', UploadAs) <= 0) and (Pos ('$', UploadAs) > 0) and (Pos ('_', UploadAs) > 0) and (CountSubStr ('_', UploadAs) = 2) then
  UploadAs:=StringReplace (UploadAs, '_', ':', [rfReplaceAll]);
 PossibleLocation:=Copy (RF, 1, X-1);
 if DirectoryExists (PossibleLocation) or FileExists (PossibleLocation) then
  begin
  X:=CountSubStr (':', UploadAs);
  if X > 0 then
   if X = 1 then
    if UploadAs [1] = ':' then
     begin
     RF:=PossibleLocation+UploadAs+':$DATA';
     VPath:=VPath+UploadAs+':$DATA';
     end
    else
     begin
     RF:=PossibleLocation+':'+UploadAs;
     VPath:=VPath+':'+UploadAs;
     end
   else
    begin
    RF:=PossibleLocation+UploadAs;
    VPath:=VPath+UploadAs;
    end
  else
   begin
   RF:=PossibleLocation+':'+UploadAs+':$DATA';
   VPath:=VPath+':'+UploadAs+':$DATA';
   end;
  end
 else
  begin
  Result:=FS_FILE_NOTFOUND;
  Exit;
  end;
 if FileExists (RF) and ((CopyFlags and FS_COPYFLAGS_OVERWRITE) = 0) then
  begin
  Move (VPath [1], RemoteName [0], Min (MAX_PATH, Length (VPath)+1));
  Result:=FS_FILE_EXISTS;
  Exit;
  end;
 if ProgressProc (Plug_ID, LocalName, PChar (RF), 0) = 1 then
  begin
  Result:=FS_FILE_USERABORT;
  Exit;
  end;
 // open files
 WS:=UNC+LF;
 hFile:=CreateFileW (PWideChar (WS), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);
 if hFile = INVALID_HANDLE_VALUE then
  begin
  Result:=FS_FILE_NOTFOUND;
  Exit;
  end;
 try
  Sz:=0;
  HiSz:=Pointer (DWord (@Sz)+4);
  SzLo:=Windows.GetFileSize (hFile, HiSz);
  if (SzLo = INVALID_FILE_SIZE) and (GetLastError <> NO_ERROR) then
   begin
   Result:=FS_FILE_READERROR;
   Exit;
   end;
  Sz:=Sz+SzLo;
  if Sz > 0 then
   hMapS:=CreateFileMapping (hFile, nil, PAGE_READONLY, HiSz^, SzLo, nil);
 finally
  CloseHandle (hFile);
 end;
 if (Sz > 0) and (hMapS = 0) then
  begin
  Result:=FS_FILE_READERROR;
  Exit;
  end;
 try
  SetLastError (0);
  WS:=UNC+RF;
  hFile:=CreateFileW (PWideChar (WS), GENERIC_READ or GENERIC_WRITE, 0, nil, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN or FILE_ATTRIBUTE_NORMAL, 0);
  if hFile = INVALID_HANDLE_VALUE then
   begin
   RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_WriteError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
   Result:=FS_FILE_OK;
   Exit;
   end;
  try
   if Sz > 0 then
    begin
    SetLastError (0);
    hMapD:=CreateFileMapping (hFile, nil, PAGE_READWRITE, HiSz^, SzLo, nil);
    end;
  finally
   CloseHandle (hFile);
  end;
  if (Sz > 0) and (hMapD = 0) then
   begin
   RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_WriteError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
   Result:=FS_FILE_OK;
   Exit;
   end;
  try
   if Sz > 0 then
    begin
    // copy data
    Sz100:=Sz div 100;
    if Sz100 <= 0 then
     Sz100:=1;
    HiOffset:=Pointer (DWord (@Offset)+4);
    Offset:=0;
    while Sz >= CopyBufferSize do
     begin
     Result:=CopyDataBlock (CopyBufferSize);
     if Result <> 0 then
      Exit;
     end;
    if Sz <> 0 then
     begin
     Result:=CopyDataBlock (Sz);
     if Result <> 0 then
      Exit;
     end;
    end;
   // close files
  finally
   if hMapD <> 0 then
    CloseHandle (hMapD);
  end;
 finally
  if hMapS <> 0 then
   CloseHandle (hMapS);
 end;
 // finish job
 if ProgressProc (Plug_ID, LocalName, PChar (RF), 100) = 1 then
  begin
  Result:=FS_FILE_USERABORT;
  Exit;
  end;
 if (CopyFlags and FS_COPYFLAGS_MOVE) <> 0 then
  if DeleteFile (LocalName) then
   Result:=FS_FILE_OK
  else
   begin
   RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_DeleteError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
   Result:=FS_FILE_WRITEERROR;
   end
 else
  Result:=FS_FILE_OK
except
 on E: Exception do
  Result:=FS_FILE_READERROR;
end;
end;

function FsRenMovFile (aOldName, aNewName: PChar; aMove, aOverWrite: Bool;
    RemoteInfo: pRemoteInfo): Integer; stdcall;

 function  CountSubStr (C: Char; const S: String): Integer;
 var
   X: Integer;

 begin
 Result:=0;
 for X:=1 to Length (S) do
  Inc (Result, Ord (S [X] = C));
 end;

var
  LF, RF, PossibleLocation, UploadAs, VPath: String;
  Sz, Sz100, Offset: Int64;
  SzLo: DWord;
  hFile, hMapS, hMapD: THandle;
  Source, Destination: Pointer;
  HiSz, HiOffset: PDWord;
  WS: WideString;
  X: Integer;

 function  CopyDataBlock (BlockSize: Integer): Integer;
 begin
 Result:=0;
 Source:=MapViewOfFile (hMapS, FILE_MAP_READ, HiOffset^, Offset, BlockSize);
 if Source = nil then
  begin
  Result:=FS_FILE_READERROR;
  Exit;
  end;
 try
  Destination:=MapViewOfFile (hMapD, FILE_MAP_WRITE, HiOffset^, Offset, BlockSize);
  if Destination = nil then
   begin
   Result:=FS_FILE_WRITEERROR;
   Exit;
   end;
  try
   Move (Source^, Destination^, BlockSize);
  finally
   UnmapViewOfFile (Destination);
  end;
 finally
  UnmapViewOfFile (Source);
 end;
 Sz:=Sz-BlockSize;
 Offset:=Offset+BlockSize;
 if ProgressProc (Plug_ID, PChar (LF), PChar (RF), Offset div Sz100) = 1 then
  begin
  Result:=FS_FILE_USERABORT;
  Exit;
  end;
 end;

begin
Result:=FS_FILE_OK;
if InitFailed then
 Exit;
try
 // prepare rename/move/copy
 RF:=aNewName;
 X:=LastDelimiter ('\', RF);
 if X <= 1 then
  begin
  RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_RootCopyUnsupported), nil, 0);
  Result:=FS_FILE_NOTSUPPORTED;
  Exit;
  end;
 VPath:=Copy (RF, 1, X);
 RF:=MakePath (RF);
 UploadAs:=Copy (RF, X+1, MaxInt);
 PossibleLocation:=Copy (RF, 1, X-1);
 if DirectoryExists (PossibleLocation) or FileExists (PossibleLocation) then
  begin
  X:=CountSubStr (':', UploadAs);
  if X > 0 then
   if X = 1 then
    if UploadAs [1] = ':' then
     begin
     RF:=PossibleLocation+UploadAs+':$DATA';
     VPath:=VPath+UploadAs+':$DATA';
     end
    else
     begin
     RF:=PossibleLocation+':'+UploadAs;
     VPath:=VPath+':'+UploadAs;
     end
   else
    begin
    RF:=PossibleLocation+UploadAs;
    VPath:=VPath+UploadAs;
    end
  else
   begin
   RF:=PossibleLocation+':'+UploadAs+':$DATA';
   VPath:=VPath+':'+UploadAs+':$DATA';
   end;
  end
 else
  begin
  Result:=FS_FILE_NOTFOUND;
  Exit;
  end;
 LF:=aOldName;
 if LastDelimiter ('\', LF) <= 1 then
  begin
  RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_RootCopyUnsupported), nil, 0);
  Result:=FS_FILE_NOTSUPPORTED;
  Exit;
  end;
 ParseStreamName (aOldName, LF);
 if AnsiLowerCase (LF) = AnsiLowerCase (RF) then
  begin
  Result:=FS_FILE_OK;
  Exit;
  end;
 if DirectoryExists (LF) then
  begin // moving/renaming directories
  if ProgressProc (Plug_ID, aOldName, aNewName, 0) = 1 then
   begin
   Result:=FS_FILE_USERABORT;
   Exit;
   end;
  if aMove then
   begin
   if DirectoryExists (RF) then
    begin
    RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_NotImplemented), nil, 0);
    Result:=FS_FILE_NOTSUPPORTED;
    Exit;
    end;
   SetLastError (0);
   if not MoveFile (PChar (LF), PChar (RF)) then
    begin
    RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_WriteError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
    Result:=FS_FILE_OK;
    Exit;
    end;
   end
  else
   begin
   RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_NotImplemented), nil, 0);
   Result:=FS_FILE_NOTSUPPORTED;
   Exit;
   end;
  if ProgressProc (Plug_ID, aOldName, aNewName, 100) = 1 then
   begin
   Result:=FS_FILE_USERABORT;
   Exit;
   end;
  Result:=FS_FILE_OK;
  Exit;
  end
 else
  if FileExists (LF) then
   begin // moving/renaming files
   if (not aMove) or (LF [1] <> RF [1]) or (CountSubStr (':', LF) > 1) or (CountSubStr (':', RF) > 1) then
    begin // source or destination is stream or files are on different drives or we need copy
    if FileExists (RF) and (not aOverWrite) then
     begin
     Move (VPath [1], aNewName [0], Min (MAX_PATH, Length (VPath)+1));
     Result:=FS_FILE_EXISTS;
     Exit;
     end;
    if ProgressProc (Plug_ID, PChar (LF), PChar (RF), 0) = 1 then
     begin
     Result:=FS_FILE_USERABORT;
     Exit;
     end;
    // open files
    WS:=UNC+LF;
    SetLastError (0);
    hFile:=CreateFileW (PWideChar (WS), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);
    if hFile = INVALID_HANDLE_VALUE then
     begin
     RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_ReadError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
     Result:=FS_FILE_OK;
     Exit;
     end;
    try
     Sz:=0;
     HiSz:=Pointer (DWord (@Sz)+4);
     SzLo:=Windows.GetFileSize (hFile, HiSz);
     if (SzLo = INVALID_FILE_SIZE) and (GetLastError <> NO_ERROR) then
      begin
      Result:=FS_FILE_READERROR;
      Exit;
      end;
     Sz:=Sz+SzLo;
     if Sz > 0 then
      begin
      SetLastError (0);
      hMapS:=CreateFileMapping (hFile, nil, PAGE_READONLY, HiSz^, SzLo, nil);
      end;
    finally
     CloseHandle (hFile);
    end;
    if (Sz > 0) and (hMapS = 0) then
     begin
     RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_ReadError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
     Result:=FS_FILE_OK;
     Exit;
     end;
    try
     SetLastError (0);
     WS:=UNC+RF;
     hFile:=CreateFileW (PWideChar (WS), GENERIC_READ or GENERIC_WRITE, 0, nil, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN or FILE_ATTRIBUTE_NORMAL, 0);
     if hFile = INVALID_HANDLE_VALUE then
      begin
      RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_WriteError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
      Result:=FS_FILE_OK;
      Exit;
      end;
     try
      if Sz > 0 then
       begin
       SetLastError (0);
       hMapD:=CreateFileMapping (hFile, nil, PAGE_READWRITE, HiSz^, SzLo, nil);
       end;
     finally
      CloseHandle (hFile);
     end;
     if (Sz > 0) and (hMapD = 0) then
      begin
      RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_WriteError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
      Result:=FS_FILE_OK;
      Exit;
      end;
     try
      if Sz > 0 then
       begin
       // copy data
       Sz100:=Sz div 100;
       if Sz100 <= 0 then
        Sz100:=1;
       HiOffset:=Pointer (DWord (@Offset)+4);
       Offset:=0;
       while Sz >= CopyBufferSize do
        begin
        Result:=CopyDataBlock (CopyBufferSize);
        if Result <> 0 then
         Exit;
        end;
       if Sz <> 0 then
        begin
        Result:=CopyDataBlock (Sz);
        if Result <> 0 then
         Exit;
        end;
       end;
      // close files
     finally
      if hMapD <> 0 then
       CloseHandle (hMapD);
     end;
    finally
     if hMapS <> 0 then
      CloseHandle (hMapS);
    end;
    // finish job
    if ProgressProc (Plug_ID, PChar (LF), PChar (RF), 100) = 1 then
     begin
     Result:=FS_FILE_USERABORT;
     Exit;
     end;
    if aMove then
     if DeleteFile (LF) then
      Result:=FS_FILE_OK
     else
      begin
      RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_DeleteError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
      Result:=FS_FILE_OK;
      end
    else
     Result:=FS_FILE_OK
    end
   else
    begin // moving usual files (not streams) inside same drive
    if ProgressProc (Plug_ID, PChar (LF), PChar (RF), 0) = 1 then
     begin
     Result:=FS_FILE_USERABORT;
     Exit;
     end;
    if FileExists (RF) then
     if aOverWrite then
      begin
      if not DeleteFile (RF) then
       begin
       RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_DeleteError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
       Result:=FS_FILE_OK;
       Exit;
       end;
      end
     else
      begin
      Result:=FS_FILE_EXISTS;
      Exit;
      end;
    if ProgressProc (Plug_ID, PChar (LF), PChar (RF), 50) = 1 then
     begin
     Result:=FS_FILE_USERABORT;
     Exit;
     end;
    SetLastError (0);
    if not MoveFile (PChar (LF), PChar (RF)) then
     begin
     RequestProc (Plug_ID, RT_MsgOK, nil, PChar (msg_WriteError+#13#10+SysErrorMessage (GetLastError)), nil, 0);
     Result:=FS_FILE_OK;
     Exit;
     end;
    if ProgressProc (Plug_ID, PChar (LF), PChar (RF), 100) = 1 then
     begin
     Result:=FS_FILE_USERABORT;
     Exit;
     end;
    Result:=FS_FILE_OK;
    end;
   end
  else
   begin
   Result:=FS_FILE_NOTFOUND;
   Exit;
   end;
except
 on E: Exception do
  Result:=FS_FILE_READERROR;
end;
end;

function FsDeleteFile (RemoteName: PChar): Bool; stdcall;
var
  S: String;

begin
if InitFailed then
 begin
 Result:=True;
 Exit;
 end;
try
 if ParseStreamName (RemoteName, S) then
  begin
  DeleteFile (S);
  Result:=True;
  Exit;
  end;
 // ***
 // (uncomment, if you don't want to remove file, while removing "::$DATA" stream. comment in another case)
 if Pos ('::', RemoteName) > 0 then
  begin
  Result:=True;
  Exit;
  end;
 // ***
 SetLastError (0);
 Result:=DeleteFile (S);
except
 on E: Exception do
  Result:=False;
end;
end;

function FsRemoveDir (RemoteName: PChar): Bool; stdcall;
var
  S: String;

begin
if InitFailed then
 begin
 Result:=True;
 Exit;
 end;
try
 if ParseStreamName (RemoteName, S) then
  begin
  Result:=True; // should be never executed (file streams are files, not directories)
  Exit;
  end;
 if FileExists (S) then
  Result:=DeleteFile (S)
 else
  Result:=DelTree (S);
except
 on E: Exception do
  Result:=False;
end;
end;

function FsSetAttr (RemoteName: PChar; NewAttr: Integer): Bool; stdcall;
var
  S: String;

begin
if InitFailed then
 begin
 Result:=True;
 Exit;
 end;
try
 if ParseStreamName (RemoteName, S) then
  begin
  Result:=True;
  Exit;
  end;
 Result:=SetFileAttributes (PChar (S), NewAttr);
except
 on E: Exception do
  Result:=False;
end;
end;

function FsSetTime (RemoteName: PChar; CreationTime, LastAccessTime,
    LastWriteTime: PFileTime): Bool; stdcall;
var
  S: String;

begin
if InitFailed then
 begin
 Result:=True;
 Exit;
 end;
try
 if ParseStreamName (RemoteName, S) then
  begin
  Result:=True;
  Exit;
  end;
 Result:=True;
 if Assigned (CreationTime) then
  Result:=Result and SetFileCreation (S, FileTimeToDateTime (CreationTime^));
 if Assigned (LastAccessTime) then
  Result:=Result and SetFileLastAccess (S, FileTimeToDateTime (LastAccessTime^));
 if Assigned (LastWriteTime) then
  Result:=Result and SetFileLastWrite (S, FileTimeToDateTime (LastWriteTime^));
except
 on E: Exception do
  Result:=False;
end;
end;

(*
procedure FsStatusInfo (RemoteDir: PChar; InfoStartEnd, InfoOperation: Integer); stdcall;
begin
// isn't used
end;
*)

procedure FsGetDefRootName (DefRootName: PChar; MaxLen: Integer); stdcall;
var
  S: String;

begin
try
 S:=Copy ('NTFS - streams', 1, MaxLen-1);
 Move (S [1], Pointer (DefRootName)^, Length (S)+1);
except
 on E: Exception do;
end;
end;

function FsExtractCustomIcon (RemoteName: PChar; ExtractFlags: Integer; var
    TheIcon: HIcon): Integer; stdcall;

 function GetIcon (Path: String): HIcon;
 var
   Ext, H: String;
   IconL, IconS: HIcon;
   IconIndex, X: Integer;

 begin
 try
  if not Assigned (Reg) then
   begin
   Result:=0;
   Exit;
   end;
  Ext:=ExtractFileExt (Path);
  if not Reg.KeyExists ('\'+Ext) then
   begin
   Result:=0;
   Exit;
   end;
  Ext:=Reg.ReadString ('\'+Ext, '', '');
  if Ext = '' then
   begin
   Result:=0;
   Exit;
   end;
  Ext:=Reg.ReadString ('\'+Ext+'\DefaultIcon', '', '');
  if Ext = '%1' then
   begin
   Ext:=Path;
   IconIndex:=0;
   end
  else
   begin
   X:=LastDelimiter (',', Ext);
   if X > 0 then
    begin
    try
     IconIndex:=StrToInt (Copy (Ext, X+1, Length (Ext)));
     if IconIndex > 0 then
      Dec (IconIndex);
    except
     on E: Exception do
      IconIndex:=0;
    end;
    Ext:=Copy (Ext, 1, X-1);
    end
   else
    IconIndex:=0;
   end;
  SetLength (H, 4097);
  SetLength (H, ExpandEnvironmentStrings (PChar (Ext), PChar (H), 4096));
  if ExtractIconEx (PChar (H), IconIndex, IconL, IconS, 1) > 0 then
   if (ExtractFlags and FS_ICONFLAG_SMALL) <> 0 then
    Result:=IconS
   else
    Result:=IconL
  else
   Result:=0;
 except
  on E: Exception do
   Result:=0;
 end;
 end;

var
  Path: String;

begin
if InitFailed then
 begin
 TheIcon:=0;
 Result:=FS_ICON_USEDEFAULT;
 Exit;
 end;
try
 if (ExtractFlags and FS_ICONFLAG_BACKGROUND) = 0 then
  begin
  TheIcon:=0;
  Result:=FS_ICON_DELAYED;
  Exit;
  end;
 if ParseStreamName (RemoteName, Path) then
  begin
  TheIcon:=LoadIcon (hInstance, 'FILESTREAM');
  Result:=FS_ICON_EXTRACTED;
  end
 else
  if FileExists (Path) then
   begin
   TheIcon:=GetIcon (Path);
   if TheIcon = 0 then
    begin
    TheIcon:=LoadIcon (hInstance, 'NULLICON');
    Result:=FS_ICON_EXTRACTED;
    end
   else
    Result:=FS_ICON_EXTRACTED_DESTROY;
   end
  else
   begin
   TheIcon:=0;
   Result:=FS_ICON_USEDEFAULT;
   end;
except
 on E: Exception do
  begin
  TheIcon:=0;
  Result:=FS_ICON_USEDEFAULT;
  end;
end;
end;

procedure FsSetDefaultParams (dps: pFsDefaultParamStruct); stdcall;
begin
try
 if dps = nil then
  Move (dps^, TC_Ver, SizeOf (TC_Ver))
 else
  ZeroMemory (@TC_Ver, SizeOf (TC_Ver));
except
 on E: Exception do;
end;
end;

{__ DLL support code __________________________________________________________}

exports
  FsInit,
  FsFindFirst,
  FsFindNext,
  FsFindClose,
  FsMkDir,
  FsExecuteFile,
  FsRenMovFile, 
  FsGetFile,
  FsPutFile,
  FsDeleteFile,
  FsRemoveDir,
  FsSetAttr,
  FsSetTime,
//  FsStatusInfo,   // notification isn't used in plugin
  FsGetDefRootName,
  FsExtractCustomIcon,
  FsSetDefaultParams;

var
  SavedDLLProc: TDLLProc;

procedure CustomDLLProc (Reason: Integer);
begin
case Reason of
 DLL_PROCESS_DETACH:
  begin
  try
   if Assigned (Reg) then
    FreeAndNil (Reg);
  except
   on E: Exception do;
  end;
  DLLProc:=SavedDLLProc;
  end;
end;
if Assigned (SavedDLLProc) then
 SavedDLLProc (Reason);
end;

begin
try
 SavedDLLProc:=DLLProc;
 DLLProc:=@CustomDLLProc;
 InitFailed:=True;
 Plug_ID:=-1;
 ZeroMemory (@TC_Ver, SizeOf (TC_Ver));
 ProgressProc:=nil;
 LogProc:=nil;
 RequestProc:=nil;
 Reg:=nil;
except
 on E: Exception do;
end;
end.
